#ifndef __EXC_H
#define __EXC_H

#if defined(_MSC_VER) && defined(_M_IX86) && (_M_IX86 >= 300) /* Visual-C */
#define __MSC32__
#endif

#if defined(__HIGHC__) || defined(__FLAT__) || \
    defined(__WATCOMC__) || defined(__MSC32__) || \
    (defined (__DMC__) && defined(DOS386))
  #define DOSX_PHARLAP
  #include <stdio.h>
  #include <pharlap.h>

#elif defined (__DJGPP__)
  #define DOSX_DJGPP
  #define UCHAR    unsigned char
  #define UINT     unsigned int
  #define USHORT   unsigned short
  #define ULONG    unsigned long
  #define NEARPTR  unsigned long
  #define REALPTR  unsigned long
  #define SWI_REGS        __dpmi_regs
  #define INT_STACK_FRAME __dpmi_regs
  typedef struct {
          ULONG   Off;
          USHORT  Sel;
        } FARPTR;

#else
  #error Unsupported compiler environment
#endif

#if defined(__cplusplus) || defined(__CPLUSPLUS__)
  extern "C" {
#endif

typedef struct {
        UINT excNumber, excCode;
        UINT DS,  ES,  FS,  GS,  SS, CS;
        UINT EIP, EAX, EBX, ECX, EDX;
        UINT ESI, EDI, EFL, ESP, EBP;
        UINT CR0, CR2;
      } excReg;

typedef struct {
        UINT     number;            /* interrupt number  */
        REALPTR  callBack;          /* callback function */
        REALPTR  rmVector;          /* old rmode vector  */
        FARPTR   pmVector;          /* old pmode vector  */
      } PMRM_INTR;

typedef struct {                    /* try/except structure (BCC32 only) */
        UINT    exceptFlag;
        UINT    exceptTOS;          /* top-of-stack in __try block       */
        USHORT  unknown[2];
        UINT    excAction;          /* __except() value or proc address  */
        void  (*exceptPoint)(void); /* __except() block start address    */
      } TRY_EXC_BLOCK;

#define NO_EXC_HOOK  (excHook)0L
typedef void (*excHook)(excReg *r);
typedef void (*SigFunc)(int);

#if !defined (__DJGPP__)
  typedef void (*pmodeHook)(SWI_REGS *);
  typedef void (*rmodeHook)(void);

  typedef void (*pmodeIntr)(INT_STACK_FRAME *);
  typedef void (*rmodeIntr)(SWI_REGS *);
  typedef void (*IrqHandler)(int, void *, SWI_REGS *);

  #define NUM_EX  18  /* HW exceptions 0-17, SW signals >= 18 */
#else
  #define NUM_EX  5   /* Signals; SIGSEGV,SIGFPE,SIGTRAP,SIGNOFP,SIGILL */
#endif

extern void  InstallExcHandler (excHook hook);
extern char *GetArg0           (USHORT sel);


#if !defined (__DJGPP__)

/* sig_1.c */

extern SigFunc  _signal (int sig, SigFunc func);
extern int      _raise  (int sig);
extern int      _kill   (int pid, int sig);
extern int      _pause  (void);
extern unsigned _alarm  (unsigned sec);

/* rmcb.c */

extern REALPTR _dx_alloc_rmode_wrapper_retf (pmodeHook pmHook,
                                             rmodeHook rmHook,
                                             int len, int stack_size);

extern REALPTR _dx_alloc_rmode_wrapper_iret (pmodeHook pmHook, int stack_size);

extern void    _dx_free_rmode_wrapper       (REALPTR dosAddr);


/* intr.c */

extern int     _dx_install_intr (rmodeIntr rmHandler,
                                 pmodeIntr pmHandler,
                                 UINT intr_num, PMRM_INTR *intr);

extern int     _dx_dump_intr    (void);

extern void    _dx_cleanup_intr (PMRM_INTR *intr);

/* reg.asm */

extern USHORT My_CS (void);
extern USHORT My_DS (void);
extern USHORT My_ES (void);
extern USHORT My_FS (void);
extern USHORT My_GS (void);
extern ULONG  My_EAX (void);
extern ULONG  My_EBX (void);
extern ULONG  My_ECX (void);
extern ULONG  My_EDX (void);
extern ULONG  My_EBP (void);
extern ULONG  My_EDI (void);
extern ULONG  My_ESI (void);
extern ULONG  My_ESP (void);

#endif  /* !__DJGPP__ */

/*
 * The following should not be used by application code
 */
#ifdef __EXC_INTERNAL

  /*
   *  Modify values (0 or 1) for features wanted
   */
  #define USERHOOK     1  /* allow user to install hook called before atexit funcs */
  #define STACKREWIND  1  /* do a stack dump with rewind of calls & args */
  #define REALLOC_SYM  0  /* realloc symbols and exc-handler to far code/data */
  #define USECOREFILE  1  /* dump to stderr or write to "core" file */
  #define DEMANGLE_CPP 1  /* demangle C++ symbols (make them readable) */
  #define AUTO_INIT    0  /* call InstallExcHandler() in constructor (djgpp only) */

  #if defined (__HIGHC__)
    #pragma Off(Align_members)  /* Pharlap forgot to pack structures */
    #include <hw386.h>          /* inside some of these headers      */
    #include <plexp.h>
    #pragma Pop(Align_members)

    #pragma Offwarn (572)
    #pragma On (relax_func_ptr_rules)
    #pragma On (Preload_args_from_memory)
    #pragma On (Recognize_library)
    #pragma On (char_is_rep)

    extern char       _end;             /* at end of BSS segment */
    extern UINT       _mwstack_limit[]; /* stack-check limits    */
    #define APP_END() (ULONG)&_end
    #define EMIT(x)    _inline(x)

  #elif defined(__MSC32__)
    #include <string.h>
    #pragma warning(disable:4103 4761)
    #pragma intrinsic (strlen)
    #pragma pack(1)
    #include <hw386.h>
    #include <plexp.h>
    #pragma pack()

    extern char        _end;
    #define APP_END()  (ULONG)&_end
    #define EMIT(x)    __emit__(x)

  #elif defined(__FLAT__) && defined(__BORLANDC__)
    #pragma pack(1)
    #include <hw386.h>
    #include <plexp.h>
    #include <string.h>
    #pragma pack()

    #define strlen  __strlen__
    #define strncpy __strncpy__
    #define __BCC32__

    extern char       _end;
    #define APP_END() (ULONG)&_end
    #define EMIT(x)    __emit__(x)

  #elif defined(__386__) && defined(__WATCOMC__)
    #undef FP_OFF
    #include <i86.h>
    #undef FP_OFF

    #pragma pack(1)
    #include <hw386.h>
    #include <plexp.h>
    #pragma pack()

    #pragma intrinsic (strlen,strcpy);

    extern char _end;
    #define APP_END() (ULONG)&_end
    #define EMIT(x)   _asm db x

  #elif defined (__DJGPP__)
    #include <unistd.h>
    #include <coff.h>
    #include <debug/syms.h>
    #include <debug/stab.h>

    extern unsigned     end __asm__ ("end");
    #define APP_END()   (ULONG)&end
    #define EMIT(x)     asm(".byte "#x"")

  #elif defined(__DMC__) && defined(DOS386)
    #pragma pack(1)
    #include <hw386.h>
    #include <plexp.h>
    #pragma pack()
    #include <syms.h>

    extern char        end;
    #define APP_END()  (ULONG)&end
    #define EMIT(x)    __emit__(x)

  #else
    #error unsupported platform
  #endif

  extern int    excAsmLines;
  extern int    excAsmDump;
  extern int    excCplusPlusSyms;
  extern int    excProgIsCplusPlus;

  extern int    excSetText;
  extern int    excCoreFile;
  extern int    excDemangleCpp;
  extern ULONG  excAppStart;
  extern ULONG  excAppEnd;
  extern ULONG  excStkStart;
  extern ULONG  excStkEnd;

  #if defined (__DJGPP__)
    #define exc_exit(x)   exit(x)
    #define exc_close(fd) _dos_close(fd)

  #else
    extern excReg context;
    extern excReg try_exc_jb;
    extern void  (*excHookMain)();

    extern FARPTR oldExcHandlers [NUM_EX];

    extern USHORT         excCseg, excDseg;
    extern UINT           exc_BOS, exc_TOS;
    extern UINT           except_TOS, except_EBP;
    extern TRY_EXC_BLOCK *excBlk;
    extern PMRM_INTR      oldintr_8, oldintr_9, oldintr_1B;
    extern PMRM_INTR      oldintr_1C,oldintr_23;

    extern void   ExcGlue_0 (void);        /* for exception 0-7,9,14-17 */
    extern void   ExcGlue_1 (void);        /* for exception 8, 10-14    */

    extern void   init_exit (void (*fn)(), UINT stk);
    extern void   exc_exit  (int status);
    extern int    exc_alloc (ULONG pages, USHORT *sel);
    extern void   exc_write (int handle, char *buf, int len);
    extern void   exc_close (int handle);

    extern USHORT Get_CS    (void);
    extern void   PmodeGlue (int taskId);
    extern void   PmGlueEnd (void);

    extern void   SigInt_8  (void);
    extern void   SigInt_9  (void);
    extern void   SigInt_8pm(INT_STACK_FRAME *reg);
    extern void   SigInt_9pm(INT_STACK_FRAME *reg);
  #endif /* __DJGPP__ */

  extern void   RemoveExcHandler (void);
  extern char  *AddressSymbol (USHORT cs, ULONG eip);
  extern void   GetCodeLimit  (ULONG *limit);
  extern void   GetDataLimit  (ULONG *limit);


  #if defined(__HIGHC__)

    /* All "internal" public symbols have a "@_" prefix
     */

    /* exc_1.c */
    pragma Alias (AddressSymbol, "@_AddressSymbol");
    pragma Alias (GetCodeLimit,  "@_GetCodeLimit");
    pragma Alias (excAsmLines,   "@_excAsmLines");
    pragma Alias (excAsmDump,    "@_excAsmDump");
    pragma Alias (excSetText,    "@_excSetText");
    pragma Alias (excCoreFile,   "@_excCoreFile");
    pragma Alias (excDemangleCpp,"@_excDemangleCpp");
    pragma Alias (excAppStart,   "@_excAppStart");
    pragma Alias (excAppEnd,     "@_excAppEnd");
    pragma Alias (excStkStart,   "@_excStkStart");
    pragma Alias (excStkEnd,     "@_excStkEnd");

    /* exc_0.asm */
    pragma Alias (excHookMain,   "@_excHookMain");
    pragma Alias (excCseg,       "@_excCseg");
    pragma Alias (excDseg,       "@_excDseg");
    pragma Alias (context,       "@_context");
    pragma Alias (exc_BOS,       "@_exc_BOS");
    pragma Alias (exc_TOS,       "@_exc_TOS");
    pragma Alias (oldExcHandlers,"@_oldExcHandlers");
    pragma Alias (ExcGlue_0,     "@_ExcGlue_0");
    pragma Alias (ExcGlue_1,     "@_ExcGlue_1");
    pragma Alias (exc_write,     "@_exc_write");
    pragma Alias (exc_close,     "@_exc_close");
    pragma Alias (init_exit,     "@_init_exit");
    pragma Alias (exc_exit,      "@_exc_exit");
    pragma Alias (exc_alloc,     "@_exc_alloc");
    pragma Alias (RemoveExcHandler,"@_RemoveExcHandler");

    /* glue.asm */
    pragma Alias (PmodeGlue, "@_PmodeGlue");
    pragma Alias (PmGlueEnd, "@_PmGlueEnd");

    /* reg.asm */
    pragma Alias (My_CS,  "@_My_CS");
    pragma Alias (My_DS,  "@_My_DS");
    pragma Alias (My_ES,  "@_My_ES");
    pragma Alias (My_FS,  "@_My_FS");
    pragma Alias (My_GS,  "@_My_GS");
    pragma Alias (My_EAX, "@_My_EAX");
    pragma Alias (My_EBX, "@_My_EBX");
    pragma Alias (My_ECX, "@_My_ECX");
    pragma Alias (My_EDX, "@_My_EDX");
    pragma Alias (My_EBP, "@_My_EBP");
    pragma Alias (My_EDI, "@_My_EDI");
    pragma Alias (My_ESI, "@_My_ESI");
    pragma Alias (My_ESP, "@_My_ESP");

    /* seh_0.asm */
    pragma Alias (excBlk,     "@_excBlk");
    pragma Alias (except_TOS, "@_except_TOS");
    pragma Alias (except_EBP, "@_except_EBP");
    pragma Alias (try_exc_jb, "@_try_exc_jb");

    /* sig_1.c */
    pragma Alias (oldintr_8,  "@_oldintr_8");
    pragma Alias (oldintr_9,  "@_oldintr_9");
    pragma Alias (oldintr_1B, "@_oldintr_1B");
    pragma Alias (oldintr_1C, "@_oldintr_1C");
    pragma Alias (oldintr_23, "@_oldintr_23");
    pragma Alias (SigInt_8,   "@_SigInt_8");
    pragma Alias (SigInt_9,   "@_SigInt_9");

    /* sig_0.asm */
    pragma Alias (SigInt_8pm, "@_SigInt_8pm");
    pragma Alias (SigInt_9pm, "@_SigInt_9pm");

    pragma Alias (_mwstack_limit, "_mwstack_limit");  /* no translation */
    pragma Alias (_end,           "_end");

    extern void far_memcpy (_Far void *dst, _Far const void *src, int n);
    pragma Alias (far_memcpy, "@_far_memcpy");

  #elif defined(__WATCOMC__)

    /* exc_0.asm */
    #pragma aux excHookMain      "@_*"
    #pragma aux excCseg          "@_*"
    #pragma aux excDseg          "@_*"
    #pragma aux context          "@_*"
    #pragma aux exc_BOS          "@_*"
    #pragma aux exc_TOS          "@_*"
    #pragma aux oldExcHandlers   "@_*"
    #pragma aux ExcGlue_0        "@_*" parm caller []
    #pragma aux ExcGlue_1        "@_*" parm caller []
    #pragma aux exc_write        "@_*" parm caller [] modify [eax ecx edx]
    #pragma aux exc_close        "@_*" parm caller [] modify [eax]
    #pragma aux init_exit        "@_*" parm caller [] modify [eax]
    #pragma aux exc_exit         "@_*" __aborts
    #pragma aux exc_alloc        "@_*" parm caller [] modify [eax ecx edx]
    #pragma aux RemoveExcHandler "@_*" parm caller [] modify [eax ecx edx]

    /* glue.asm */
    #pragma aux PmodeGlue        "@_*"
    #pragma aux PmGlueEnd        "@_*"

    /* reg.asm */
    #pragma aux My_CS            "@_*" parm caller [] modify [eax]
    #pragma aux My_DS            "@_*" parm caller [] modify [eax]
    #pragma aux My_ES            "@_*" parm caller [] modify [eax]
    #pragma aux My_FS            "@_*" parm caller [] modify [eax]
    #pragma aux My_GS            "@_*" parm caller [] modify [eax]
    #pragma aux My_EAX           "@_*" parm caller [] modify [eax]
    #pragma aux My_EBX           "@_*" parm caller [] modify [eax]
    #pragma aux My_ECX           "@_*" parm caller [] modify [eax]
    #pragma aux My_EDX           "@_*" parm caller [] modify [eax]
    #pragma aux My_EBP           "@_*" parm caller [] modify [eax]
    #pragma aux My_EDI           "@_*" parm caller [] modify [eax]
    #pragma aux My_ESI           "@_*" parm caller [] modify [eax]
    #pragma aux My_ESP           "@_*" parm caller [] modify [eax]

    /* seh_0.asm */
    #pragma aux excBlk           "@_*"
    #pragma aux except_TOS       "@_*"
    #pragma aux except_EBP       "@_*"
    #pragma aux try_exc_jb       "@_*"

    /* sig_0.asm */
    #pragma aux SigInt_8pm       "@_*"
    #pragma aux SigInt_9pm       "@_*"
    #pragma aux SigInt_8         "@_*"
    #pragma aux SigInt_9         "@_*"
    #pragma aux oldintr_8        "@_*"
    #pragma aux oldintr_9        "@_*"

    extern void far_memcpy (void far *dst, const void far *src, int n);
    #pragma aux far_memcpy       "@_*" parm caller [] modify [eax ecx]
  #endif

#endif  /* __EXC_INTERNAL */

#if defined(__cplusplus) || defined(__CPLUSPLUS__)
  };
#endif

#endif  
