#ifndef __EXC_C
#error This file should be included from ../exc_1.c
#endif

#undef  STDOUT_FILENO
#undef  STDERR_FILENO
#define STDOUT_FILENO  1
#define STDERR_FILENO  2

/*
 * Most functions are called at exception time. Turn off
 * stack-checking to prevent reentrancy. HighC generates INT0
 * on stack-overflow.
 */

#if defined(__HIGHC__)
  #pragma Off (check_stack)
#elif defined(__MSC32__)
  #pragma check_stack (off)
#elif defined(__BCC32__)
  #pragma option -N-
#endif


static char *excName[] = {
            "Divide Exception", "Debug Exception",       /* 0  */
            "NMI !!??",         "Breakpoint",            /* 2  */
            "Overflow",         "Array Bound Check",     /* 4  */
            "Invalid Opcode",   "FPU not present",       /* 6  */
            "Double Fault",     "FPU Segment Overrun",   /* 8  */
            "Invalid TSS",      "Segment not present",   /* 10 */
            "Stack Fault",      "General Protection",    /* 12 */
            "Page Fault",       "Reserved !!??",         /* 14 */
            "FPU Error",        "Alignment Check",       /* 16 */
            "Timer signal",     "^C pressed",     /* internals */
            "^Break pressed"
          };

/*
 * Set exception handler for exception `x'.
 */
static void HookExcHandler (int x, USHORT cs)
{
  FARPTR new;

  /* BREAK/DEBUG exception needs special handling. If debugger is
   * active we must not replace exc1/exc3 with our handler.
   */
  if (x == INT_INT3 || x == INT_DBG)
  {
    FARPTR iv, ex;

    _dx_pmiv_get (INT_PROTF, &iv);
    _dx_excep_get(INT_PROTF, &ex);
    if (FP_SEL(iv) != FP_SEL(ex)) /* separate exc-13 gates ? */
       return;                    /* yes, debugger active    */
  }
  FP_SET (new, ExcHasErrCode(x) ? &ExcGlue_1 : &ExcGlue_0, cs);
  _dx_excep_set (x, new);
}

/*
 *  Application interface to exception handler system
 *  Called once during main start. Add/remove handlers as
 *  required. Can be invoked multiple times
 */
void InstallExcHandler (excHook hook)
{
  static int been_here = 0;
  CONFIG_INF cnf;
  int        i, vmm;

  if (been_here)
     return;

  been_here = 1;
  _dx_config_inf (&cnf, (UCHAR*)&cnf);  /* get config block */
  excCseg     = cnf.c_cs_sel;
  excDseg     = cnf.c_ds_sel;
  vmm         = cnf.c_vmmf && !(cnf.c_flags1 & CF1_NOVM);
  excAppStart = FindAppStart (excCseg);

#if 0
  {
    _dx_page_lockn (ExcDataStart, data_len);
    _dx_page_lockn (ExcCodeStart, code_len);
  }
#endif

#ifdef __HIGHC__
  if (excStkEnd == 0xFFFFFFFF)      /* stack limits not set by app */
      excStkEnd = _mwstack_limit[1];
#endif
  if (excAppEnd == 0xFFFFFFFF)
     GetCodeLimit (&excAppEnd);

  /* Save vectors for all exception handlers
   */
  for (i = 0; i < NUM_EX; i++)
     _dx_excep_get (i, &oldExcHandlers[i]);

  HookExcHandler ( 0, excCseg);    /* divide error exception   */
  HookExcHandler ( 1, excCseg);    /* debug exception          */
  HookExcHandler ( 3, excCseg);    /* break point exception    */
  HookExcHandler ( 4, excCseg);    /* integer overflow exc     */
  HookExcHandler ( 5, excCseg);    /* array bound exception    */
  HookExcHandler ( 6, excCseg);    /* Illegal opcode exception */
  HookExcHandler ( 7, excCseg);    /* Coprocessor not present  */
  HookExcHandler (10, excCseg);    /* invalid TSS              */
  HookExcHandler (11, excCseg);    /* segment not present exc  */
  HookExcHandler (12, excCseg);    /* stack fault exception    */
  HookExcHandler (13, excCseg);    /* general protection exc   */
  if (!vmm)                            /* if not using 386|VMM */
     HookExcHandler (14, excCseg);     /* page fault exception */

#if STACKREWIND
  InitStackTrace (GetArg0(cnf.c_env_sel));
#endif

  excHookUser = hook;
  excHookMain = FaultingFinish;
  videoMode   = PeekRealByte (0x449);
  videoHeight = PeekRealByte (0x484);  /* assume EGA+ */
  init_exit (exit,(UINT)&i);

#if REALLOC_SYM
  {
    ULONG  size = GetSymbolsSizes (NULL);
    FARPTR fp;
    USHORT sel;
    if (exc_alloc((size+4095)/4096,&sel))
    {
      FP_SET (fp,0,sel);
      ReallocPubSyms (fp,size);
#if 0  // !! to-do
      _dx_umap_pgsn (ExcDataStart, data_len);
      _dx_umap_pgsn (ExcCodeStart, code_len);
#endif
    }
  }
#endif
  atexit (RemoveExcHandler);
}


/*
 *  Since there's no standard way for getting at "char *argv[]"
 *  passed to "main()", we must do it ourself
 */
char *GetArg0 (USHORT env_sel)
{
  static  char buf[128];
  UINT    limit;
  CD_DES  descr;
  FARPTR  fp;
  char    *env, *start;

  if (_dx_ldt_rd(env_sel,(UCHAR*)&descr))
     return (NULL);

  limit = descr.limit0_15 + ((descr.limit16_19 & 15) << 16);

  if ((env = malloc(limit)) == NULL)
     return (NULL);

  start = env;
  FP_SET (fp,0,env_sel);
  ReadFarMem ((void*)env,fp,limit);

  /* The environment is organised like this:
   * 'path=c:\dos;c:\util',0
   *  ..
   * 'comspec=c:\dos\command.com',0,0,0
   * 'program.exe',0
   *
   */
  while (*(env+1))
    env += 1 + strlen(env);
  env += 2;
  strncpy (buf, env, sizeof(buf));
  free (start);
  return (buf);
}


/*
 * Find the first allocated page. I.e find the parameter
 * of the "-offset" link parameter
 */
STATIC UINT FindAppStart (USHORT sel)
{
  FARPTR fp;

#ifdef __HIGHC__
  _FP_SEG (fp) = sel;
  _FP_OFF (fp) = 0;
#else
  FP_SET (fp, 0, sel);
#endif

  while (1)
  {
    ULONG pte = 0;  /* Page table entry */
    ULONG pti = 0;  /* Page table info  */

    if (_dx_rd_ptinf(fp,&pte,&pti) == _DXE_NOT_DPMI || pte)
       break;
    FP_INCR (fp, 4096);
  }
#ifdef __WATCOMC__
  return (unsigned)fp;
#else
  return FP_OFF (fp);
#endif
}

/*
 * Return to text-mode if video-mode changed since we
 * installed ourself. Also check for 50-line mode.
 */
static void SetTextMode (void)
{
  if (PeekRealByte(0x449) != videoMode)
  {
    SWI_REGS reg;

    reg.eax = 0x83;             /* set mode 3, no clear   */
    _dx_real_int (0x10, &reg);
    if (videoHeight > 25)       /* was 50-line screen     */
    {
      reg.eax = 0x1112;         /* load 8x8 font, block 0 */
      reg.ebx = 0;
      _dx_real_int (0x10, &reg);
    }
  }
}

/*
 * Return segment limits of CS and DS selectors
 */
void GetCodeLimit (ULONG *limit)
{
  CD_DES descr;

  if (_dx_ldt_rd(SS_CODE,(UCHAR*)&descr) == 0)
       *limit = (descr.limit0_15 + ((descr.limit16_19 & 15) << 16)) << 14;
  else *limit = excStkStart;
}

void GetDataLimit (ULONG *limit)
{
  CD_DES descr;

  if (_dx_ldt_rd(SS_DATA,(UCHAR*)&descr) == 0)
       *limit = (descr.limit0_15 + ((descr.limit16_19 & 15) << 16)) << 14;
  else *limit = APP_END();
}

