PAGE  60,132
TITLE Pmode callback glue
NAME  callback_glue

INCLUDE EXC.INC

_DATA SEGMENT

TASKRECORD struc
           swi_unused dd  0,0    ; | SWI_REGS (386asm dosn't support struct
           swi_eax    dd  0      ; | within struct
           swi_ebx    dd  0      ; |
           swi_ecx    dd  0      ; |
           swi_edx    dd  0      ; |
           swi_esi    dd  0      ; |
           swi_edi    dd  0      ; |
           swi_ebp    dd  0      ; |
           swi_esp    dd  0      ; |
           swi_cs     dw  0      ; |
           swi_ds     dw  0      ; |
           swi_ss     dw  0      ; |
           swi_es     dw  0      ; |
           swi_fs     dw  0      ; |
           swi_gs     dw  0      ; |
           swi_eip    dd  0      ; |
           swi_eflags dd  0      ; |
           dosReal    dd  0
           stackStart dd  0
           stackTop   dd  0
           callback   dd  0
           marker     dd  0DEADBEEFh
TASKRECORD ends

extrn @rmcb : TASKRECORD

_DATA ENDS

_TEXT SEGMENT


; void PmodeGlue (char taskIndex)
;
; We need an assembly language "glue" routine on the protected mode side
; to deal with several things:
;	1.  It gets called with a FAR procedure call
;       2.  Copy all real-mode registers into the rmReg structure of the
;           correct taskId
;       3.  Switch back to a stack in the data segment (32-bit C code
;           requires SS and DS to be the same).
;       4.  Call the HLL prot-mode callback with &rmReg on stack
;       5.  Switch back to the stack DosX has allocated for us.
;           (this stack maps the real-mode stack we where called with)
;       6.  Copy rmRegs back onto the stack

C_PROC PmodeGlue
       cld                             ; C code need direction flag off

; Switch stacks.  We can freely use the general registers here, since they
; are all saved in the real mode handler.

       mov  ebx, [esp+8]               ; EBX = taskIndex
       imul ebx, SIZE TASKRECORD
       lea  ebx, @rmcb [ebx]           ; EBX -> rmcb-data [taskIndex]

       call FillRmRegsFromStack        ; fill rmReg [ebx]
       mov  dx, ss
       mov  esi, esp                   ; DX:ESI = current SS:ESP
       mov  ax,ds
       mov  ss,ax
       mov  esp, [ebx].stackTop        ; new SS:ESP = DS:top-of-task-stack
       push edx                        ; save old stack pointer on top
       push esi                        ; of new stack

       push ebx                        ; because rmReg is first in the structure
       call [ebx].callback             ; call the C handler function
       lss  esp, pword ptr [esp+4]     ; switch back to original stack
       jmp  short FillStackFromRmReg   ; fill stack with rmReg (ebx unchanged)


;-------------------------------------------------------------------------
; fill the rmReg struct based on the "pushad" instruction and the
; "push ds/es" of the wrapper. Registers that haven't changed are stored
; as-is

TOP = 4                                ; top of stack where registers are

FillRmRegsFromStack:
       mov [ebx].swi_eax, eax    ; EAX
       mov [ebx].swi_ecx, ecx    ; ECX
       mov [ebx].swi_edx, edx    ; EDX
       mov [ebx].swi_esi, esi    ; ESI
       mov [ebx].swi_edi, edi    ; EDI

       pushfd
       pop eax
       mov [ebx].swi_eflags, eax ; EFLAGS

       mov eax, [esp+TOP+28]
       mov [ebx].swi_ebx, eax    ; EBX

       mov eax, [esp+TOP+24]
       add eax, 4*4 + 3*2
       mov [ebx].swi_esp, eax    ; ESP

       mov eax, [esp+TOP+20]
       mov [ebx].swi_ebp, eax    ; EBP

       mov eax, [esp+TOP+44]
       mov [ebx].swi_ds, ax      ; DS

       mov eax, [esp+TOP+46]
       mov [ebx].swi_es, ax      ; ES

       mov eax, [esp+TOP+48]
       mov [ebx].swi_ss, ax      ; SS
       ret

;-------------------------------------------------------------------------
; fill stack with the rmReg struct based on the "popad" instruction
; and the "pop ds/es" of the wrapper. We don't reload SS,ESP or EBP

TOP = 0

FillStackFromRmReg:
       mov ax, [ebx].swi_ss      ; SS
       mov [esp+TOP+48], ax

       mov ax, [ebx].swi_es      ; ES
       mov [esp+TOP+46], ax

       mov ax, [ebx].swi_ds      ; DS
       mov [esp+TOP+44], ax

       mov eax, [ebx].swi_eax    ; EAX
       mov [esp+TOP+40], eax

       mov eax, [ebx].swi_ecx    ; ECX
       mov [esp+TOP+36], eax

       mov eax, [ebx].swi_edx    ; EDX
       mov [esp+TOP+32], eax

       mov eax, [ebx].swi_ebx    ; EBX
       mov [esp+TOP+28], eax

       mov eax, [ebx].swi_esi    ; ESI
       mov [esp+TOP+16], eax

       mov eax, [ebx].swi_edi    ; EDI
       mov [esp+TOP+12], eax
ifdef WATCOM
       retf                      ; return to real mode code
else
       RetFar     
endif

C_PROC PmGlueEnd                 ; end of area to lock


_TEXT ENDS
END
