/*
 *  Structured exception handling (SEH) support for PharLap 386|DosX
 *  Only for Borland's BCC32 and Microsoft VC/MSC.
 *
 *  by G. Vanem 10-Oct-1996
 */

#if defined(__BORLANDC__) || defined(__MSC32__)

#include <stdio.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winnt.h>

#define __EXC_INTERNAL
#include "exc.h"

#ifdef __BORLANDC__
#pragma warn -pro
#endif

#ifndef cdecl
#define cdecl
#endif

#define jmp_buf       excReg      /* define jmp_buf our own way */
#define setjmp(x)    _exc_setjmp
#define longjmp(x,y) _exc_longjmp(&(x),y)

extern jmp_buf _try_exc_jb;

extern cdecl void _exc_longjmp  (jmp_buf *buf, int retval);
extern cdecl int  _exc_setjmp   (jmp_buf *buf);
extern cdecl void _set_filt_args(int exc_code, EXCEPTION_POINTERS *info);

static EXCEPTION_POINTERS * SetExceptionInformaton (int *code);

/*
 * Exception handler HLL entry for SEH.
 * exc_0 calls this hook (at ExcCont:). We never return to exc_0,
 * but longjmp to (try/)except block(s)
 */
void except_try_hook (void)
{
  long (*filter)();
  long  action = excBlk->excAction;
  int   exc_code;

  /* copy the exception context (from exc_0) to local jmp_buf
   */
  memcpy (&_try_exc_jb,&context,sizeof(context));

  _set_filt_args (exc_code, SetExceptionInformaton(&exc_code));

  do
  {
    if (action == EXCEPTION_EXECUTE_HANDLER)
    {
      _try_exc_jb.EIP = (unsigned int) excBlk->exceptPoint;
      break;
    }
    if (action == EXCEPTION_CONTINUE_EXECUTION)
    {
      (*excBlk->exceptPoint)();
      break;
    }
    if (action == EXCEPTION_CONTINUE_SEARCH)
    {
      /* to-do, search next "exception" handler or "finally" block
       */
      break;
    }

    /* call __except() filter routine
     */
    filter = (long(*)()) action;
    action = (*filter)();
    continue;
  }
  while (0);

  /*
   * BCC32 emits a "mov fs:[0],x" near the try block ending.
   * This trips a page-fault if compiled with "-offset 1000h".
   * PSP selector is always 0-based; therefore set FS = SS_PSP
   */
  _try_exc_jb.FS  = SS_PSP;
  _try_exc_jb.ESP = _try_exc_jb.EBP = except_TOS;
  longjmp (_try_exc_jb,0);
}

/*
 * Translate exc-numbers to codes defined in <winnt.h>
 */
static EXCEPTION_POINTERS * SetExceptionInformaton (int *exc_code)
{
  static unsigned long excCodeTab [NUM_EX] = {
         STATUS_INTEGER_DIVIDE_BY_ZERO,     /* 0 : Divide Exception     */
         STATUS_SINGLE_STEP,                /*   : Debug Exception      */
         0,                                 /* 2 : NMI !!??             */
         STATUS_BREAKPOINT,                 /*   : Breakpoint           */
         STATUS_INTEGER_OVERFLOW,           /* 4 : Overflow             */
         STATUS_ARRAY_BOUNDS_EXCEEDED,      /*   : Array Bound Check    */
         STATUS_ILLEGAL_INSTRUCTION,        /* 6 : Invalid Opcode       */
         STATUS_FLOAT_INVALID_OPERATION,    /*   : FPU not present      */
         0,                                 /* 8 : Double Fault         */
         STATUS_FLOAT_OVERFLOW,             /*   : FPU Segment Overrun  */
         STATUS_INVALID_DISPOSITION,        /* 10: Invalid TSS          */
         STATUS_ACCESS_VIOLATION,           /*   : Segment not present  */
         STATUS_STACK_OVERFLOW,             /* 12: Stack Fault          */
         STATUS_ACCESS_VIOLATION,           /*   : General Protection   */
         STATUS_IN_PAGE_ERROR,              /* 14: Page Fault           */
         STATUS_FLOAT_INVALID_OPERATION,    /*   : FPU Error            */
         STATUS_DATATYPE_MISALIGNMENT       /* 16: Alignment Check"     */
       };

  static  EXCEPTION_POINTERS       exc;       /* exception information  */
  static  struct _CONTEXT          ctx;
  static  struct _EXCEPTION_RECORD excRec;
  static  FLOATING_SAVE_AREA       floatSave; /* float status, not init */
  static  SYS_REGS                 sysReg;    /* DosX system regs; DRx  */

  _dx_sysregs_get (&sysReg);

  ctx.ContextFlags = CONTEXT_CONTROL  | CONTEXT_INTEGER  |
                     CONTEXT_SEGMENTS | CONTEXT_DEBUG_REGISTERS;

  ctx.FloatSave = floatSave;
  ctx.Eip       = context.EIP;
  ctx.SegCs     = context.CS;
  ctx.SegFs     = context.FS;
  ctx.SegGs     = context.GS;
  ctx.SegDs     = context.DS;
  ctx.SegSs     = context.SS;
  ctx.Edi       = context.EDI;
  ctx.Esi       = context.ESI;
  ctx.Ebx       = context.EBX;
  ctx.Edx       = context.EDX;
  ctx.Ecx       = context.ECX;
  ctx.Eax       = context.EAX;
  ctx.Ebp       = context.EBP;
  ctx.Esp       = context.ESP;
  ctx.EFlags    = context.EFL;
  ctx.Dr0       = sysReg.dr0;
  ctx.Dr1       = sysReg.dr1;
  ctx.Dr2       = sysReg.dr2;
  ctx.Dr3       = sysReg.dr3;
  ctx.Dr6       = sysReg.dr6;
  ctx.Dr7       = sysReg.dr7;

  excRec.ExceptionCode    = excCodeTab [context.excNumber];
  excRec.ExceptionFlags   = 0;
  excRec.ExceptionRecord  = NULL;     /* no next exception */
  excRec.ExceptionAddress = (PVOID) context.EIP;
  excRec.NumberParameters = 0;

  exc.ContextRecord   = &ctx;
  exc.ExceptionRecord = &excRec;
  *exc_code = excRec.ExceptionCode;
  return (&exc);
}

#else

  void except_try_hook (void) {} /* to keep make+linker happy */

#endif /* __BORLANDC__ || __MSC32__ */
