#ifndef __TRACE_H
#define __TRACE_H

#ifndef ULONG
#define ULONG  unsigned long
#endif

#ifndef USHORT
#define USHORT unsigned short
#endif

#ifndef UINT
#define UINT   unsigned int
#endif

#ifndef UCHAR
#define UCHAR  unsigned char
#endif

#if defined(__DJGPP__)

typedef struct SYM_ENTRY {
        ULONG   string_off;
        UCHAR   type;
        UCHAR   other;
        USHORT  desc;
        ULONG   val;
      } SYM_ENTRY;

typedef struct FileNode {
        char    *filename;
        ULONG    first_address;
        ULONG    last_address;
        LINENO  *lines;
        int      num_lines;
      } FileNode;

#else

#include "packon.h"

typedef struct SymHeader {
        char   name [4];            /* "SYM1" or "SYM2"             */
        USHORT headSize;            /* 1Eh                          */
        ULONG  segSymSize;          /* size of segment symbol table */
        ULONG  pubSymSize;          /* size of public symbol table  */
      } SymHeader;

typedef struct CVsymHeader {
        char   name [4];            /* "SYM2"                       */
        USHORT headSize;            /*                              */
        ULONG  segSymSize;          /* size of segment symbol table */
        ULONG  pubSymSize;          /* size of public symbol table  */
      } CVsymHeader;

#include "packoff.h"

#endif  /* !__DJGPP__ */


typedef struct _SymNode  {          /* public symbol structure      */
        struct _SymNode *next;      /* next-node pointer            */
        char            *name;      /* symbol ASCIIZ name           */
        ULONG           ofs;        /* symbol address: offset       */
        USHORT          sel;        /*                 selector     */
        USHORT          line;       /*                 line/file    */
        char            file[8];
      } SymNode;

typedef struct _SegNode {           /* segment symbol structure     */
     // struct _SegNode *next;      /* next-node pointer            */
        char            *name;      /* symbol ASCIIZ name           */
        ULONG           ofs;        /* selector offset in image     */
        ULONG           limit;      /* initial selector limit       */
        USHORT          selector;   /* selector value               */
      } SegNode;


#define SIGN_MZ  0x5A4D             /* "MZ" */
#define SIGN_MQ  0x514D             /* "MQ" */
#define SIGN_P2  0x334F             /* "P2" */
#define SIGN_P3  0x3350             /* "P3" */

#define MAX_SEL       4
#define MAX_SYM_NAME  100

#if defined(__DJGPP__)
  ULONG GetCoffSymbols (int fd, long ofs);             /* in coff.c */
  ULONG GetAoutSymbols (int fd, long ofs);             /* in aout.c */
#else
  ULONG GetPharLapSymbols (char *buf, SymHeader *sh);  /* in pharlap.c */
  ULONG GetCodeViewSymbols(char *buf, SymHeader *sh);  /* in cview.c */
  UINT  ReallocPubSyms  (FARPTR area, ULONG max);
#endif

char    *InitStackTrace  (char *name);
SymNode *AddAddrSymbol   (char *name, ULONG ofs, USHORT sel);
SegNode *AddSegSymbol    (char *name, ULONG ofs, ULONG sel, ULONG limit);
SymNode *FindPublicSym   (ULONG ofs);
char    *GetSymbolStr    (ULONG eip);
void     RewindStack     (ULONG start, ULONG base, int fd);
ULONG    GetSymbolsSizes (ULONG *num);
USHORT   GetSelectorVal  (USHORT idx);

#ifdef __HIGHC__
  #define MIN(a,b) _min(a,b)
  #define MAX(a,b) _max(a,b)
#else
  #define MIN(a,b) ((a) <= (b) ? (a) : (b))
  #define MAX(a,b) ((a) >  (b) ? (a) : (b))
#endif

#ifndef ARGSUSED
#define ARGSUSED(x) ((x) = (x))
#endif

#endif

